/*
 * Copyright (C) 1996-2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"

/***
 *** 3D display for robotics applications
 ***/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <tcl.h>
#include <tk.h>
#include <togl.h>

#include "gdhe.h"
#include "global.h"
#include "cb.h"
#include "commands.h"
#include "light.h"
#include "objets.h"
#include "server.h"

static int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */
static Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
static Tcl_DString line;	/* Used to read the next line from the
                                 * terminal input. */

static int nosetup = 0;			/* don't call the setup proc on 
					 * startup */

static void Prompt(Tcl_Interp *interp, int partial);
static void StdinProc(ClientData clientData, int mask);
static void gdheError(Tcl_Interp *interp, char *msg);
static int gdheToglInit(Tcl_Interp *interp);
static int TclDefProcInit(Tcl_Interp *interp);
static int TclVarInit(Tcl_Interp *interp);
static Tcl_Interp *interp;	/* Interpreter for this application. */

/* Une fonction interne de Tk, bien pratique cependant */
extern void TkpDisplayWarning (char *msg, char *title);

/*----------------------------------------------------------------------*/

/**
 ** fonction d'initialisation de GDHE
 **/

int
main(int argc, char * const argv[])
{
    char *args, *fileName, *logname = NULL;
    Tcl_Channel outChannel, inChannel;
    char buf[1024];
    int code, c, errflag = 0;
    struct rlimit limit;
    int tkinit = 0;

    /* Evite la production de core (gigantesques a cause de OpenGL) */
    if (getrlimit(RLIMIT_CORE, &limit) == -1) {
	perror("GDHE: getrlimit");
	exit(1);
    }
    limit.rlim_cur = 0;
    if (setrlimit(RLIMIT_CORE, &limit) == -1) {
	perror("GDHE: limit coredumpsize 0");
	exit(1);
    }

    /* code pomp de tkMain.c */
    Tcl_FindExecutable(argv[0]);
    tclInterp = interp = Tcl_CreateInterp();

    while ((c = getopt(argc, argv, "l:qv")) != EOF) {
	switch (c) {
	  case 'l':
	    logname = strdup(optarg);
	    break;
	case 'q':
		nosetup = 1;
		break;
	case 'v':
		printf("GDHE version " GDHE_VERSION "\n");
		exit(0);
	  case '?':
	    errflag++;
	} /* switch */
    } /* while */
    if (errflag) {
	gdheError(interp, "GDHE: usage: gdhe [-l logfile][-v]");
    }

    fileName = NULL;
    if ((argc > 1) && (argv[1][0] != '-')) {
        fileName = argv[1];
        argc--;
        argv++;
    }

    args = Tcl_Merge(argc-1, (const char * const *)argv+1);
    Tcl_SetVar(interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    snprintf(buf, sizeof(buf), "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0",  (fileName != NULL) ? fileName : argv[0],
	       TCL_GLOBAL_ONLY);

    tty = isatty(0);
    Tcl_SetVar(interp, "tcl_interactive",
            ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /* Initialise Tcl et Tk */
    if (Tcl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

    if (Tk_Init(interp) == TCL_ERROR) {
	/* if a filename was given, still source it before failing. This let
           pkg_mkIndex be built from the Makefile without an X11 connection */
      fputs("cannot initialize tk\n", stderr);
      if (!fileName) {
	return TCL_ERROR;
      }
    } else tkinit = 1;

    /* Initialise le reste des variables d'environnement de Tcl */
    TclVarInit(interp);

    if (gdheToglInit(interp) != TCL_OK) {
	    gdheError(interp, "GDHE: Togl initialisation failed");
    }

    if (fileName != NULL) {
        code = Tcl_EvalFile(interp, fileName);
        if (code != TCL_OK) {
            /*
             * The following statement guarantees that the errorInfo
             * variable is set properly.
             */

	     Tcl_AddErrorInfo(interp, "");
	     fputs(Tcl_GetVar(interp, "errorInfo",
		     TCL_GLOBAL_ONLY), stderr);
            Tcl_DeleteInterp(interp);
            Tcl_Exit(1);
        }
        tty = 0;
    } else {

	 /* source les fichiers de ressources */
	 if (!nosetup && getenv("GDHE_NO_SETUP") == NULL) {
		    TclDefProcInit(interp);
	 }

	 if (init_xdr(logname) != 0) {
		 return TCL_ERROR;
	 }
	 /*
	  * Establish a channel handler for stdin.
	  */

	 inChannel = Tcl_GetStdChannel(TCL_STDIN);
	 if (inChannel) {
	      Tcl_CreateChannelHandler(inChannel, TCL_READABLE, StdinProc,
				       (ClientData) inChannel);
	 }
	 if (tty) {
	      Prompt(interp, 0);
	 }
    }
    outChannel = Tcl_GetStdChannel(TCL_STDOUT);
    if (outChannel) {
	 Tcl_Flush(outChannel);
    }
    Tcl_ResetResult(interp);

    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

    if (!tkinit) return TCL_OK;
    Tk_MainLoop();

    Tcl_DeleteInterp(interp);
    Tcl_Exit(0);
    return(0);

}

/**
 ** fonction d'initialisation de GDHE (package)
 **/

int
Gdhe_Init(Tcl_Interp *interp)
{
     int code;
     char *logname = NULL;

     tclInterp = interp;
     TclVarInit(interp);

     code = Tcl_PkgProvide(interp, "Gdhe", "1.2");
     if (code != TCL_OK) {
	     return code;
     }

     printf("loading Gdhe Tcl module\n");

     if (gdheToglInit(interp) != TCL_OK) {
	  gdheError(interp, "GDHE: Togl initialisation failed");
     }
     if (getenv("GDHE_NO_SETUP") == NULL) {
	     TclDefProcInit(interp);
     

	     /* Cree le socket et le fichier de log */
	     logname = getenv("GDHE_LOGNAME");
	     if (init_xdr(logname) != 0) {
		     return TCL_ERROR;
	     }
     }
     return TCL_OK;
}

static int
TclVarInit(Tcl_Interp *interp)
{
    /* Note: GDHE_DATADIR and GDHE_MODULESDIR are not defined in
     * config.h, but given in the command line because
     * $prefix and $exec_prefix are unusable in configure.ac
     */
    Tcl_SetVar(interp, "gdhe_datadir", GDHE_DATADIR, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "gdhe_library", GDHE_DATADIR "/tcl", TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "gdhe_modulesdir", GDHE_MODULESDIR, TCL_GLOBAL_ONLY);

    /* Ajoute gdhe_library dans auto_path */
    Tcl_Eval(interp, "lappend auto_path " GDHE_DATADIR "/tcl" );
    Tcl_Eval(interp, "lappend auto_path " GDHE_MODULESDIR);
}

static int
TclDefProcInit(Tcl_Interp *interp)
{
    Tcl_Obj *postRcCmd;

     /* charge les fonctions */
     if (Tcl_Eval(interp, "source $gdhe_library/setup.tcl") != TCL_OK) {
	  gdheError(interp, "GDHE: error loading gdhe library");
     }

     /* Defini le fichier de config... */
     Tcl_SetVar(interp, "tcl_rcFileName", ".gdherc", TCL_GLOBAL_ONLY);

     /* ...et execute-le */
     Tcl_SourceRCFile(interp);

     /* Enfin, lance l'appli */
     if (Tcl_Eval(interp, "setup") != TCL_OK) {
	  gdheError(interp, "GDHE: error excuting setup");
     }
     /* Look for the "gdhe_after_setup_rc" variable and evaluate it */
     postRcCmd = Tcl_ObjGetVar2(interp,
				Tcl_NewStringObj("gdhe_after_setup_rc", -1),
				NULL, TCL_GLOBAL_ONLY);
     if (postRcCmd != NULL) {
	 if (Tcl_EvalObjEx(interp, postRcCmd, TCL_EVAL_DIRECT|TCL_EVAL_GLOBAL)
	     != TCL_OK) {
	     gdheError(interp, "GHE: error evaluating gdhe_after_setup_rc");
	 }
     }
     return(0);
}

static void
gdheError(Tcl_Interp *interp, char *msg)
{
    Tcl_Channel errChannel;
    const char *tclError;

    errChannel = Tcl_GetStdChannel(TCL_STDERR);
    if (errChannel) {

	Tcl_Write(errChannel, msg, -1);
	Tcl_Write(errChannel, ": ", 2);
	Tcl_Write(errChannel, Tcl_GetStringResult(interp), -1);
	Tcl_Write(errChannel, "\n", 1);

	tclError = Tcl_GetVar(interp, "errorInfo", 0);
	if (tclError != NULL) {
	    Tcl_Write(errChannel, tclError, -1);
	    Tcl_Write(errChannel, "\n", 1);
	}
    }
    exit(2);
}

static int
gdheToglInit(Tcl_Interp *interp)
{
    const char *s;

    /* Initialise Togl */
    if (Togl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }
    /* Creation du nullObj */
    nullObj = Tcl_NewStringObj("", -1);
    Tcl_IncrRefCount(nullObj);

    /*
     * Installe les callbacks
     */

    /* callbacks ToGL */
    Togl_CreateFunc(create_cb);
    Togl_DisplayFunc(display_cb);
    Togl_ReshapeFunc(reshape_cb);

    /* Installe les commandes Togl */
    Togl_CreateCommand("redraw", redrawCallBack);
    Togl_CreateCommand("setObs", setObsCallBack);
    Togl_CreateCommand("polygonMode", polygonModeCallBack);
    Togl_CreateCommand("dumpEps", dumpEpsCallBack);
    Togl_CreateCommand("dumpPpm", dumpPpmCallBack);

    /* de objects.c */
    Tcl_CreateObjCommand(interp, "box", boxCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "prism", prismCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "cylinder", cylinderCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "polygon", polygonCallBack, NULL, NULL);
#ifdef notyet
    Tcl_CreateObjCommand(interp, "transparentPolygon", transparentPolygonCallBack, NULL, NULL);
#endif
    Tcl_CreateObjCommand(interp, "polyline", polylineCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "reedShep", reedShepCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "sphere", sphereCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "ellipsoid", ellipsoidCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "disk", diskCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "circle", circleCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "repere", repereCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "drawString", drawStringCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "picture", pictureCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "concave", concaveCallBack, NULL, NULL);

    /* de commands.c */
    Tcl_CreateObjCommand(interp, "color", colorCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "clearColor", clearColorCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "material", materialCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "translate", translateCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "rotate", rotateCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "pushMatrix", pushMatrixCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "loadMatrix", loadMatrixCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "loadIdentity", loadIdentityCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "multMatrix", multMatrixCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "popMatrix", popMatrixCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "getMVMatrix", getMVMatrixCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "genLists", genListsCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "deleteLists", deleteListsCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "callList", callListCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "newList", newListCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "endList", endListCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "cullFace", cullFaceCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "setLights", setLightsCallBack, NULL, NULL);
    /* XXX */
    Tcl_CreateObjCommand(interp, "enableLighting", enableLightingCallBack,NULL, NULL);
    Tcl_CreateObjCommand(interp, "enableLightingNow", enableLightingNowCallBack,  NULL, NULL);
    Tcl_CreateObjCommand(interp, "disableLightingNow", disableLightingNowCallBack, NULL, NULL);

    Tcl_CreateObjCommand(interp, "sleep", sleepCallBack, NULL, NULL);

    Tcl_CreateObjCommand(interp, "drawString",drawStringCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "annotate", frameAnnotateCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "initNames", initNamesCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "popName", popNameCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "pushName", pushNameCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "loadName", loadNameCallBack, NULL, NULL);
    Togl_CreateCommand("pickUp", pickUpCallBack);

    /* Liaison des variables globales */
    gdheGetDefaultFromInterp(interp, "xmin", &xmin, xmin);
    gdheGetDefaultFromInterp(interp, "xmax", &xmax, xmax);
    gdheGetDefaultFromInterp(interp, "ymin", &ymin, ymin);
    gdheGetDefaultFromInterp(interp, "ymax", &ymax, ymax);
    gdheGetDefaultFromInterp(interp, "zmin", &zmin, zmin);
    gdheGetDefaultFromInterp(interp, "zmax", &zmax, zmax);

    Tcl_LinkVar(interp, "xmin", (char *)&xmin, TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp, "ymin", (char *)&ymin, TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp, "zmin", (char *)&zmin, TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp, "xmax", (char *)&xmax, TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp, "ymax", (char *)&ymax, TCL_LINK_DOUBLE);
    Tcl_LinkVar(interp, "zmax", (char *)&zmax, TCL_LINK_DOUBLE);

    Tcl_LinkVar(interp, "enable_lighting", (char *)&enable_lighting,
		TCL_LINK_INT);

    Tcl_LinkVar(interp, "auto_redisplay",
		(char *)&auto_redisplay, TCL_LINK_INT);

    /* Trace des tableaux d'eclairage */
    Tcl_TraceVar(interp, "::Gdhe::lights::position",
		 TCL_GLOBAL_ONLY|TCL_TRACE_WRITES, setLightCallback, NULL);
    Tcl_TraceVar(interp, "::Gdhe::lights::ambient",
		 TCL_GLOBAL_ONLY|TCL_TRACE_WRITES, setLightCallback, NULL);
    Tcl_TraceVar(interp, "::Gdhe::lights::diffuse",
		 TCL_GLOBAL_ONLY|TCL_TRACE_WRITES, setLightCallback, NULL);
    Tcl_TraceVar(interp, "::Gdhe::lights::enabled",
		 TCL_GLOBAL_ONLY|TCL_TRACE_WRITES, setLightCallback, NULL);

    /* Creation de la H-table des positions */
    posTable = (Tcl_HashTable *)Tcl_Alloc(sizeof(Tcl_HashTable));
    if (posTable == NULL) {
	return TCL_ERROR;
    }
    Tcl_InitHashTable(posTable, TCL_STRING_KEYS);

    return TCL_OK;
}

/*----------------------------------------------------------------------*/
/*
 *----------------------------------------------------------------------
 *
 * StdinProc --
 *
 *	This procedure is invoked by the event dispatcher whenever
 *	standard input becomes readable.  It grabs the next line of
 *	input characters, adds them to a command being assembled, and
 *	executes the command if it's complete.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Could be almost arbitrary, depending on the command that's
 *	typed.
 *
 *----------------------------------------------------------------------
 */

/* ARGSUSED */
static void
StdinProc(clientData, mask)
    ClientData clientData;		/* Not used. */
    int mask;				/* Not used. */
{
    static int gotPartial = 0;
    char *cmd;
    int code, count;
    Tcl_Channel chan = (Tcl_Channel) clientData;

    count = Tcl_Gets(chan, &line);

    if (count < 0) {
	if (!gotPartial) {
	    if (tty) {
		Tcl_Exit(0);
	    } else {
		Tcl_DeleteChannelHandler(chan, StdinProc, (ClientData) chan);
	    }
	    return;
	} else {
	    count = 0;
	}
    }

    if (count > 0) {
	(void) Tcl_DStringAppend(&command, Tcl_DStringValue(&line), -1);
    }
    cmd = Tcl_DStringAppend(&command, "\n", -1);
    Tcl_DStringFree(&line);

    if (!Tcl_CommandComplete(cmd)) {
        gotPartial = 1;
        goto prompt;
    }
    gotPartial = 0;

    /*
     * Disable the stdin channel handler while evaluating the command;
     * otherwise if the command re-enters the event loop we might
     * process commands from stdin before the current command is
     * finished.  Among other things, this will trash the text of the
     * command being evaluated.
     */

    Tcl_CreateChannelHandler(chan, 0, StdinProc, (ClientData) chan);
    code = Tcl_RecordAndEval(interp, cmd, TCL_EVAL_GLOBAL);
    Tcl_CreateChannelHandler(chan, TCL_READABLE, StdinProc,
			     (ClientData) chan);
    Tcl_DStringFree(&command);
    if (Tcl_GetStringResult(interp)[0] != 0) {
	 if ((code != TCL_OK) || (tty)) {
	    /*
	     * The statement below used to call "printf", but that resulted
	     * in core dumps under Solaris 2.3 if the result was very long.
             *
             * NOTE: This probably will not work under Windows either.
	     */
	    puts(Tcl_GetStringResult(interp));
	}
    }

    /*
     * Output a prompt.
     */

    prompt:
    if (tty) {
	 Prompt(interp, gotPartial);
    }
    Tcl_ResetResult(interp);
}

/*
 *----------------------------------------------------------------------
 *
 * Prompt --
 *
 *	Issue a prompt on standard output, or invoke a script
 *	to issue the prompt.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A prompt gets output, and a Tcl script may be evaluated
 *	in interp.
 *
 *----------------------------------------------------------------------
 */

static void
Prompt(interp, partial)
    Tcl_Interp *interp;			/* Interpreter to use for prompting. */
    int partial;			/* Non-zero means there already
					 * exists a partial command, so use
					 * the secondary prompt. */
{
    const char *promptCmd;
    int code;
    Tcl_Channel outChannel, errChannel;

    errChannel = Tcl_GetChannel(interp, "stderr", NULL);

    promptCmd = Tcl_GetVar(interp,
	partial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) {
defaultPrompt:
	if (!partial) {

            /*
             * We must check that outChannel is a real channel - it
             * is possible that someone has transferred stdout out of
             * this interpreter with "interp transfer".
             */

	    outChannel = Tcl_GetChannel(interp, "stdout", NULL);
            if (outChannel != (Tcl_Channel) NULL) {
                Tcl_Write(outChannel, "% ", 2);
            }
	}
    } else {
	code = Tcl_Eval(interp, promptCmd);
	if (code != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
		    "\n    (script that generates prompt)");
            /*
             * We must check that errChannel is a real channel - it
             * is possible that someone has transferred stderr out of
             * this interpreter with "interp transfer".
             */

	    errChannel = Tcl_GetChannel(interp, "stderr", NULL);
            if (errChannel != (Tcl_Channel) NULL) {
		Tcl_Write(errChannel, Tcl_GetStringResult(interp), -1);
                Tcl_Write(errChannel, "\n", 1);
            }
	    goto defaultPrompt;
	}
    }
    outChannel = Tcl_GetChannel(interp, "stdout", NULL);
    if (outChannel != (Tcl_Channel) NULL) {
        Tcl_Flush(outChannel);
    }
}

/*----------------------------------------------------------------------*/
/**
 ** Get the default value for a C variable from a corresponding Tcl
 ** variable, or from a default value specified in argument
 **/
void
gdheGetDefaultFromInterp(Tcl_Interp *interp, char *name,
    double *ptr, double value)
{
	Tcl_Obj *obj;

	obj = Tcl_ObjGetVar2(interp, Tcl_NewStringObj(name, -1),
	    NULL, TCL_GLOBAL_ONLY);
	if (obj == NULL ||
	    Tcl_GetDoubleFromObj(interp, obj, ptr) != TCL_OK) {
		*ptr = value;
	}
}
